""" 
Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


API Triggers: "CreateDistribution", "CreateDistributionWithTags" and "UpdateDistribution" 
Services: Amazon Cloudwatch EventBridge (trigger), Amazon SNS, AWS Lambda, AWS CloudFormation and S3
Python 3.6 - Lambda - Last Modified 02/01/2021
"""

import logging
import os
import boto3
from botocore.exceptions import ClientError

OUTBOUND_TOPIC_ARN = os.environ["outbound_topic_arn"]


def lambda_handler(event, context):
    """
    Lambda handler.

    Evaluates the control and makes messaging decisions.
    """
    setup_logging()
    log.info('Got an event!')
    log.info(event)

    tls_violations = check_tls_1_2(event)
    listener_tls_violations = check_tls_1_2_listener(event)
    https_violations = check_behavior_httpsonly(event)
    https_origin_violations = check_origin_httpsonly(event)
    logging_violations = check_logging(event)
    if "Violation" in [tls_violations, listener_tls_violations,
                       https_violations, https_origin_violations,
                       logging_violations]:
        log.info("Violations Found - Sending SNS notification")
        create_non_compliance_message(tls_violations, listener_tls_violations,
                                      https_violations, https_origin_violations,
                                      logging_violations, event, context)
    else:
        log.info("No Violations Found - Check Complete")


def check_tls_1_2(distribution_description):
    """
    Check TLS Version.

    Check origin configurations for TLS versions.
    Non compliant if anything other than TLS 1.2 is found.
    """
    violation = ""
    try:
        for item in distribution_description["detail"]["responseElements"] \
                            ["distribution"]["distributionConfig"]["origins"]["items"]:
            for config in item:
                if "customOriginConfig" in config:
                    for protocol in item["customOriginConfig"]["originSslProtocols"]["items"]:
                        if protocol != "TLSv1.2":
                            log.info("Found Origin configured for TLS that is " +
                                  "not version 1.2")
                            violation = "Violation"
                            return violation
                        else:
                            pass
        return violation
    except KeyError:
        log.info('No distribution found in the event.')


def check_tls_1_2_listener(distribution_description):
    """
    Check TLS Listener.

    Check Cloudfront listener for TLS version.
    Non compliant if anything other than TLS 1.2 is found.
    """
    listenerviolation = ""
    try:
        certificateconfig = distribution_description["detail"]["responseElements"][
            "distribution"]["distributionConfig"]["viewerCertificate"]
        if "sSLSupportMethod" in certificateconfig:
            if certificateconfig["sSLSupportMethod"] == "sni-only":
                if certificateconfig["minimumProtocolVersion"] != "TLSv1.2_2018":
                    log.info("Found Cloudfront certificate installed with TLS " +
                          "version that is not version 1.2")
                    listenerviolation = "Violation"
                    return listenerviolation
        else:
            return listenerviolation
    except KeyError:
        log.info('No distribution found in the event.')


def check_behavior_httpsonly(distribution_description):
    """
    Check for HTTPS Only.

    Check behavior configurations for HTTPS Only.
    Non compliant if port 80 is allowed.
    """
    httpsviolation = ""
    # Check default distribution config for "https-only"
    try:
        if distribution_description["detail"]["responseElements"]["distribution"]["distributionConfig"]["defaultCacheBehavior"]["viewerProtocolPolicy"] != "https-only":
            log.info("Found Default Behavior configured for something other " +
                  "than HTTPS Only")
            httpsviolation = "Violation"
    except KeyError:
        log.info('No distribution found in the event.')
    # Loop to check individual CF behaviors for their viewerProtocolPolicy
    try:
        if "items" in distribution_description["detail"]["responseElements"]["distribution"]["distributionConfig"]["cacheBehaviors"]:
            for item in distribution_description["detail"]["responseElements"]["distribution"]["distributionConfig"]["cacheBehaviors"]["items"]:
                for config in item:
                    if "viewerProtocolPolicy" in config:
                        if item["viewerProtocolPolicy"] != "https-only":
                            log.info(
                                "Found Custom Behavior configured for something " +
                                "other than HTTPS Only")
                            httpsviolation = "Violation"
                            return httpsviolation
                        else:
                            pass
        return httpsviolation
    except KeyError:
        log.info('No distribution found in the event.')


def check_origin_httpsonly(distribution_description):
    """
    Check origin for HTTPS.

    Check origin configurations for HTTPS Only.
    Non compliant if port 80 is allowed.
    """
    # Loop to check individual CF origins for their originProtocolPolicy
    httpsoriginviolation = ""
    try:
        if "items" in distribution_description["detail"]["responseElements"]["distribution"]["distributionConfig"]["origins"]:
            for item in distribution_description["detail"]["responseElements"]["distribution"]["distributionConfig"]["origins"]["items"]:
                for config in item:
                    if "customOriginConfig" in config:
                        if item["customOriginConfig"][
                                "originProtocolPolicy"] != "https-only":
                            log.info(
                                "Found Origin configured for something other " +
                                "than HTTPS Only")
                            httpsoriginviolation = "Violation"
                            return httpsoriginviolation
                    else:
                        pass
        return httpsoriginviolation
    except KeyError:
        log.info('No distribution found in the event.')


def check_logging(distribution_description):
    """
    Check logging config.

    Non compliant if logging is disabled.
    """
    loggingviolation = ""
    # Check default distribution config for logging set to disabled
    if not distribution_description["detail"]["responseElements"]["distribution"]["distributionConfig"]["logging"]["enabled"]:
        log.info("Found Logging for distribution set to Disabled")
        loggingviolation = "Violation"
        return loggingviolation
    return loggingviolation


def create_non_compliance_message(tlsviolations, listenertlsviolations,
                                  httpsviolations, httpsoriginviolations,
                                  loggingviolations, event, context):
    """
    Create Non-Compliance Message.

    Creates the subject line and body of the SNS message depending on which
    violations were encountered.
    """
    subject = "Violation - Cloudfront distribution is out of compliance!"
    message = ""
    if tlsviolations == "Violation":
        message += "Violation - Non TLS1.2 protocols are in use for distribution ID: " + \
            event["detail"]["responseElements"]["distribution"]["id"] + \
            " custom origins\n\n"
    if listenertlsviolations == "Violation":
        message += "Violation - Cloudfront custom certificate installed for distribution ID " + \
            event["detail"]["responseElements"]["distribution"]["id"] + \
            " with TLS version lower than 1.2\n\n"
    if httpsviolations == "Violation":
        message += "Violation - A Cloudfront Behavior in distribution ID: " + \
            event["detail"]["responseElements"]["distribution"]["id"] + \
            " is not HTTPS only\n\n"
    if httpsoriginviolations == "Violation":
        message += "Violation - A Cloudfront Origin in distribution ID " + \
            event["detail"]["responseElements"]["distribution"]["id"] + \
            " is not HTTPS only\n\n"
    if loggingviolations == "Violation":
        message += "Violation - Logging for distribution ID " + \
            event["detail"]["responseElements"]["distribution"]["id"] + \
            " is disabled\n\n"
    send_violation(OUTBOUND_TOPIC_ARN, message, subject, event, context)


def send_violation(outbound_topic_arn, message, subject, event, context):
    """
    Send Violation.

    Sends a notice to the specified topic arn.
    """
    findsnsregion = outbound_topic_arn.split(":")
    snsregion = findsnsregion[3]
    message += "Cloudfront Domain Name: " + \
        event["detail"]["responseElements"]["distribution"]["domainName"]
    message += '\nAccount: ' + event["account"] + "\n"
    message += "\n\n"
    message += "This notification was generated by the Lambda function " + \
        context.invoked_function_arn
    sendclient = boto3.client('sns', region_name=snsregion)
    try:
        sendclient.publish(
            TopicArn=outbound_topic_arn,
            Message=message,
            Subject=subject
        )
    except ClientError as err:
        log.info(err)
        return False


def setup_logging():
    """
    Logging Function.

    Creates a global log object and sets its level.
    """
    global log
    log = logging.getLogger()
    log_levels = {'INFO': 20, 'WARNING': 30, 'ERROR': 40}

    if 'logging_level' in os.environ:
        log_level = os.environ['logging_level'].upper()
        if log_level in log_levels:
            log.setLevel(log_levels[log_level])
        else:
            log.setLevel(log_levels['ERROR'])
            log.error("The logging_level environment variable is not set to INFO, WARNING, or \
                      ERROR.  The log level is set to ERROR")
    else:
        log.setLevel(log_levels['ERROR'])
        log.warning('The logging_level environment variable is not set. The log level is set to \
                    ERROR')
    log.info('Logging setup complete - set to log level ' + str(log.getEffectiveLevel()))